From: Masami Hiramatsu (Google) <[email protected]>

With the introduction of container_of-style BTF typecasting and
per-CPU variable access support in trace probes, we need a way to
verify their functionality and prevent regressions.

Add a new ftrace kselftest and update the trace event sample module
to test and validate these features.

Specifically, update the trace-events-sample module to set up a
periodic timer whose callback accesses a per-CPU counter. Introduce
a new sample trace event, foo_timer_fn, to trace this callback
and log the current counter value.

Then, add a new test case, btf_probe_event.tc, which defines a
dynamic probe on the timer callback. The probe uses BTF typecasting
to recover the parent structure from the timer argument and
this_cpu_read() to fetch the per-CPU counter. The test verifies
the integrity of the implementation by ensuring the values
recorded by the dynamic probe match those from the static tracepoint.

Assisted-by: Antigravity:gemini-3.5-flash
Signed-off-by: Masami Hiramatsu (Google) <[email protected]>
---
 Changes in v2:
  - Use timer_shutdown_sync() instead of timer_delete_sync() for teardown.
---
 samples/trace_events/trace-events-sample.c         |   40 +++++++++++++++-
 samples/trace_events/trace-events-sample.h         |   34 ++++++++++++-
 .../ftrace/test.d/dynevent/btf_probe_event.tc      |   51 ++++++++++++++++++++
 3 files changed, 120 insertions(+), 5 deletions(-)
 create mode 100644 
tools/testing/selftests/ftrace/test.d/dynevent/btf_probe_event.tc

diff --git a/samples/trace_events/trace-events-sample.c 
b/samples/trace_events/trace-events-sample.c
index b61766864b54..2a1f73533a38 100644
--- a/samples/trace_events/trace-events-sample.c
+++ b/samples/trace_events/trace-events-sample.c
@@ -93,6 +93,20 @@ static int simple_thread_fn(void *arg)
 
 static DEFINE_MUTEX(thread_mutex);
 
+static struct foo_timer_data *foo_timer_data;
+
+static void sample_timer_cb(struct timer_list *t)
+{
+       struct foo_timer_data *data = container_of(t, struct foo_timer_data, 
timer);
+
+       get_cpu();
+       trace_foo_timer_fn(data);
+       (*this_cpu_ptr(data->counter))++;
+       put_cpu();
+
+       mod_timer(t, jiffies + HZ);
+}
+
 int foo_bar_reg(void)
 {
        mutex_lock(&thread_mutex);
@@ -124,9 +138,27 @@ void foo_bar_unreg(void)
 
 static int __init trace_event_init(void)
 {
+       foo_timer_data = kzalloc_obj(*foo_timer_data, GFP_KERNEL);
+       if (!foo_timer_data)
+               return -ENOMEM;
+
+       foo_timer_data->name = "sample_timer_counter";
+       foo_timer_data->counter = alloc_percpu(int);
+       if (!foo_timer_data->counter) {
+               kfree(foo_timer_data);
+               return -ENOMEM;
+       }
+
+       timer_setup(&foo_timer_data->timer, sample_timer_cb, 0);
+       mod_timer(&foo_timer_data->timer, jiffies + HZ);
+
        simple_tsk = kthread_run(simple_thread, NULL, "event-sample");
-       if (IS_ERR(simple_tsk))
-               return -1;
+       if (IS_ERR(simple_tsk)) {
+               timer_shutdown_sync(&foo_timer_data->timer);
+               free_percpu(foo_timer_data->counter);
+               kfree(foo_timer_data);
+               return PTR_ERR(simple_tsk);
+       }
 
        return 0;
 }
@@ -139,6 +171,10 @@ static void __exit trace_event_exit(void)
                kthread_stop(simple_tsk_fn);
        simple_tsk_fn = NULL;
        mutex_unlock(&thread_mutex);
+
+       timer_shutdown_sync(&foo_timer_data->timer);
+       free_percpu(foo_timer_data->counter);
+       kfree(foo_timer_data);
 }
 
 module_init(trace_event_init);
diff --git a/tools/testing/selftests/ftrace/test.d/dynevent/btf_probe_event.tc 
b/tools/testing/selftests/ftrace/test.d/dynevent/btf_probe_event.tc
new file mode 100644
index 000000000000..96791e120b7d
--- /dev/null
+++ b/tools/testing/selftests/ftrace/test.d/dynevent/btf_probe_event.tc
@@ -0,0 +1,51 @@
+#!/bin/sh
+# SPDX-License-Identifier: GPL-2.0
+# description: BTF event with typecast and percpu access
+# requires: dynamic_events "this_cpu_read(<fetcharg>)":README 
"[(structname[,field])]<argname>[->field[->field|.field...]]":README
+
+# Check if the sample module is loaded
+if ! lsmod | grep -q trace_events_sample; then
+  modprobe trace-events-sample || exit_unsupported
+fi
+
+echo 0 > events/enable
+echo > dynamic_events
+
+# The sample_timer_cb(struct timer_list *t) is called.
+# We want to check (STRUCT,FIELD)VAR typecast and this_cpu_read() access.
+# (foo_timer_data,timer)t converts t to struct foo_timer_data * using 
container_of.
+# data->counter is a per-cpu pointer to int.
+# this_cpu_read(data->counter) should give the value of the counter.
+
+echo 'f:mysample/myevent sample_timer_cb 
name=(foo_timer_data,timer)t->name:string 
count=this_cpu_read((foo_timer_data,timer)t->counter)' >> dynamic_events
+
+echo 1 > events/mysample/myevent/enable
+echo 1 > events/sample-trace/foo_timer_fn/enable
+
+sleep 2
+
+echo 0 > events/mysample/myevent/enable
+echo 0 > events/sample-trace/foo_timer_fn/enable
+
+# Compare the values.
+MATCH=0
+while read line; do
+  if echo $line | grep -q "foo_timer_fn:"; then
+    NAME=`echo $line | sed 's/.*name=\([^ ]*\) .*/\1/'`
+    COUNT=`echo $line | sed 's/.*count=\([^ ]*\).*/\1/'`
+    if grep -q "myevent:.*name=\"${NAME}\" count=$COUNT" trace; then
+       MATCH=$((MATCH+1))
+    fi
+  fi
+done < trace
+
+if [ $MATCH -eq 0 ]; then
+  echo "No matching events found"
+  exit_fail
+fi
+
+# Clean up
+echo 0 > events/mysample/myevent/enable
+echo 0 > events/sample-trace/foo_timer_fn/enable
+echo > dynamic_events
+clear_trace


Reply via email to
diff --git a/samples/trace_events/trace-events-sample.h 
b/samples/trace_events/trace-events-sample.h
index 1a05fc153353..816848a456a2 100644
--- a/samples/trace_events/trace-events-sample.h
+++ b/samples/trace_events/trace-events-sample.h
@@ -247,12 +247,14 @@
  */
 
 /*
  * It is OK to have helper functions in the file, but they need to be protected
  * from being defined more than once. Remember, this file gets included more
  * than once.
  */
 #ifndef __TRACE_EVENT_SAMPLE_HELPER_FUNCTIONS
 #define __TRACE_EVENT_SAMPLE_HELPER_FUNCTIONS
+#include <linux/timer.h>
+
 static inline int __length_of(const int *list)
 {
        int i;
@@ -270,6 +272,13 @@ enum {
        TRACE_SAMPLE_BAR = 4,
        TRACE_SAMPLE_ZOO = 8,
 };
+
+struct foo_timer_data {
+       const char              *name;
+       struct timer_list       timer;
+       int __percpu            *counter;
+};
+
 #endif
 
 /*
@@ -595,6 +604,25 @@ TRACE_EVENT(foo_rel_loc,
                  __get_rel_bitmask(bitmask),
                  __get_rel_cpumask(cpumask))
 );
+
+TRACE_EVENT(foo_timer_fn,
+
+       TP_PROTO(struct foo_timer_data *data),
+
+       TP_ARGS(data),
+
+       TP_STRUCT__entry(
+               __string(       name,                   data->name      )
+               __field(        int,                    count           )
+       ),
+
+       TP_fast_assign(
+               __assign_str(name);
+               __entry->count  = *this_cpu_ptr(data->counter);
+       ),
+
+       TP_printk("name=%s count=%d", __get_str(name), __entry->count)
+);
 #endif
 
 /***** NOTICE! The #if protection ends here. *****/
@@ -247,12 +247,14 @@
 
 /*
- * It is OK to have helper functions in the file, but they need to be protected
- * from being defined more than once. Remember, this file gets included more
- * than once.
+ * It is OK to have helper functions and data structures in the file, but they
+ * need to be protected from being defined more than once. Remember, this file
+ * gets included more than once.
  */
 #ifndef __TRACE_EVENT_SAMPLE_HELPER_FUNCTIONS